import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.awt.geom.*;
import java.math.*;
import java.util.*;
import java.awt.font.*;

public class ChordCanvas extends Canvas implements MouseListener, MouseMotionListener {
    static final int MODE_BOTH = 0;
    static final int MODE_CHORD = 1;
    static final int MODE_POINT = 2;
    int PADDING = 15;
    int SIZE;
    double SCALEFACTOR;
    double CX, CY;
    float PIXEL;
    AffineTransform T, TINV;
    double UL[];
    static final double PTSIZE_PXL = 2.5;
    static final Ellipse2D.Double unitCircle = new Ellipse2D.Double(-1,-1,2.0,2.0);
    //    static final double sqrt3 = 1.73205080756887729352;
    //    static final Complex V1 = new Complex(0.0,1.0);
    //    static final Complex V2 = new Complex(-sqrt3/2.0,-0.5);
    //    static final Complex V3 = new Complex(sqrt3/2.0,-0.5);
    static final double FUZZ = 0.01;

    final IdealTriangle tri = new IdealTriangle(0.0,2.0 * Math.PI/3.0, 4.0*Math.PI/3.0);

    DrawList<TriangleChord> chords;
    double laccum = 0.0;

    Image buffer;
    Image fixedbg;
    Graphics gbuffer;

    int displayMode = MODE_BOTH;

    int alpha = 50;
    Color chordColor = new Color(100,255,100,alpha);
    Color geodesicColor = new Color(100,100,255,alpha);
    Font warningFont = new Font("SansSerif",Font.BOLD,24);

    public ChordCanvas(int size, DrawList<TriangleChord> chords) {
	this.chords = chords;

	SIZE = size;
	setSize(SIZE,SIZE);

	// Compute transformation
	SCALEFACTOR = 0.5 * SIZE - PADDING;
	CX = 0.5 * SIZE;
	CY = 0.5 * SIZE;
	PIXEL = (float)(1.0 / SCALEFACTOR);
	T = new AffineTransform();
	T.translate(CX,CY);
	T.scale(SCALEFACTOR,-1.0*SCALEFACTOR);
	try {
	    TINV = T.createInverse();
	}
	catch(Exception e) {
	    // Very bad... scale + translate was not invertible?
	    // In that case make TINV the identity.
	    TINV = new AffineTransform();
	}

	addMouseListener(this);
	addMouseMotionListener(this);
    }

    // You can't create images in the constructor!
    // Instead they are created when the canvas is added, through this
    // addNotify hook.
    public void addNotify() {
	super.addNotify();
	// Create fixed background
	fixedbg = createImage(SIZE,SIZE);
	paintFixed((Graphics2D)fixedbg.getGraphics());

	// Create buffer
	buffer = createImage(SIZE,SIZE);
	gbuffer = buffer.getGraphics();
    }

    private Shape makePoint(Complex z,double r)
    {
	return(new Ellipse2D.Double(z.re()-r*PIXEL,
				    z.im()-r*PIXEL,
				    2.0*r*PIXEL,2.0*r*PIXEL));
    }

    private Shape makePoint(Complex z)
    {
	return makePoint(z,PTSIZE_PXL);
    }

    public static Shape makeArrowHead(Complex center, Complex dir)
    {
	GeneralPath p = new GeneralPath();

	Complex z = center.add(dir);
	p.moveTo((float)z.re(),(float)z.im());
	z = center.add(Poincare.OMEGA.mul(dir));
	p.lineTo((float)z.re(),(float)z.im());
	z = center.add(Poincare.OMEGA.conj().mul(dir));
	p.lineTo((float)z.re(),(float)z.im());
	p.closePath();
	return p;
    }

    public void update(Graphics g) {
	paint(g);
    }

    public void paint(Graphics g) {
	if (chords.needRedraw() && (chords.size() > 50)) {
	    drawWarning(g);
	}
        paintReal(gbuffer);
        g.drawImage(buffer, 0, 0, this);
    }

    public void paintFixed(Graphics2D g) {
	// Pretty rendering
	g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
	// Clear
	g.setColor(Color.black);
	g.fillRect(0,0,getWidth(),getHeight());

	// From now on, use complex coordinates
	AffineTransform saved = g.getTransform();
	g.transform(T);

	// Unit circle
	g.setColor(new Color(100,100,100));
	g.setStroke(new BasicStroke((float)1.25*PIXEL));
	g.draw(unitCircle);

	// Ideal triangle
       	g.setColor(Color.white);
	tri.drawOn(g);

	// Revert to device coordinates
	g.setTransform(saved);
    }

    public void paintReal(Graphics g2) {
	//	Graphics gb2 = fixedbg.getGraphics();
	Graphics2D g = (Graphics2D)g2;
	if (chords.needRedraw()) {
	    g2.drawImage(fixedbg, 0, 0, this);
	    //	    paintfixed(g);
	}

	if (chords.hasUndrawn()) {
	    // Pretty rendering
	    g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
	    // From now on, use complex coordinates
	    AffineTransform saved = g.getTransform();
	    g.transform(T);	
    
	    // Draw all the chords in green
	    g.setStroke(new BasicStroke((float)1.0*PIXEL));
	    //	    chords.setRedraw();
	    while (chords.hasUndrawn()) {
		TriangleChord c = (TriangleChord)chords.getUndrawn();
		//		System.out.println("Now drawing item "+chords.indexOf(c)+" with center x="+c.getBase().re());
		if (displayMode < 1) {
		    g.setColor(geodesicColor);
		    c.drawGeodesicOn(g);
		}
		g.setColor(chordColor);
		if (displayMode < 2) {
		    c.drawChordOn(g);
		}
		g.fill(makePoint(c.getBase(),0.5*PTSIZE_PXL));
	    }
	
	    // Revert to device coordinates
	    g.setTransform(saved);
	}
    }

    public void addRandomChord()
    {
	Complex p, pideal; //, qideal;
	//	PoincareSegment c;
	TriangleChord c;
	p = Poincare.randomStdIdeal();
	//       	p = new Complex(0.0,0.0);
	pideal = Poincare.randomIdeal(p);

	//	c = tri.getChord(new PoincareSegment(pideal,PoincareSegment.getEndpoint(pideal,p)));
	c = new TriangleChord(tri,p,pideal);
	chords.push(c);
	//PoincareSegment(c.getIdealEnd(),c.getIdealStart()));
	//	chords.addElement(c);
	//points.addElement(p);
	//double l =  Poincare.dist(c.getStart(),c.getEnd());
	//lengths.addElement(l);
	laccum += c.getLength();
	//	System.out.println("New chord with length l=" + l);
	//	System.out.print("Added random chord centered at: ");
	//	p.println();
    }

    public void addRandomChords(int n)
    {
	for (int i=0; i<n; i++)
	    addRandomChord();
	repaint(20);
    }

    public void redraw()
    {
	chords.setRedraw();
	repaint(20);
    }

    public void setDisplayMode(int m)
    {
	displayMode = (m % 3);
	chords.setRedraw();
    }

    public void cycleDisplayMode()
    {
	setDisplayMode(displayMode + 1);
	repaint(20);
    }

    public void clearChords()
    {
	chords.clear();
	repaint(20);
    }

    private Color adjustAlpha(Color c, int alpha)
    {
	return new Color(c.getRed(), c.getGreen(), c.getBlue(), alpha);
    }

    public void setAlpha(int alpha)
    {
	this.alpha = alpha;
	chordColor = adjustAlpha(chordColor,alpha);
	geodesicColor = adjustAlpha(geodesicColor,alpha);
	redraw();
    }

    public int getAlpha()
    {
	return alpha;
    }

    public void drawWarning(Graphics g2)
    {
	Graphics2D g = (Graphics2D)g2;
	g.setColor(Color.RED);
	String s = "Drawing...";
	//	FontRenderContext frc = g.getFontRenderContext();
	//	Font f = g.getFont();
	//	Rectangle2D extent = f.getStringBounds(s,frc);
	g.setFont(warningFont);
	g.drawString("Drawing...",20,100);
	//	g.translate(20,20);
	//	g.draw(extent);
    }

    public void showAverage()
    {
	System.out.println("Current average length="+(laccum / chords.size()));
    }

    public Complex pixelToComplex(int x, int y)
    {
	double[] input = {(double)x,(double)y};
	double[] output = new double[2];
	TINV.transform(input,0,output,0,1);
	return new Complex(output[0],output[1]);
    }

    public void mouseClicked(MouseEvent e)
    {
    }

    public void mouseEntered(MouseEvent e)
    {
    }

    public void mouseExited(MouseEvent e)
    {
    }

    public void mousePressed(MouseEvent e)
    {
    }

    public void mouseReleased(MouseEvent e)
    {
// 	if (e.getButton() == 1) {
// 	    addRandomChords(50);
// 	} else if (e.getButton() == 2) {
// 	    clearChords();
// 	} else {
// 	    cycleDisplayMode();
// 	}
    }

    public void mouseMoved(MouseEvent e)
    {
    }

    public void mouseDragged(MouseEvent e)
    {
	//	if ((e.getModifiersEx() & InputEvent.BUTTON1_DOWN_MASK) == InputEvent.BUTTON1_DOWN_MASK) {
	    // Dragged while mouse button 1 is pressed... do nothing
	    //Complex pos = pixelToComplex(e.getX(),e.getY());
	    //Complex disp = pos.sub(z);
	//	}
    }
}
